(* file: part2.ml
 * author: Bob Muller
 * date: January 1, 2015
 * revised: September 8, 2015
 *
 * Various simple graphics problems for part 2 of problem set 2.
 *
 * depends: Asai/Uehara Universe Library, 
 *          Cs1101 Library
 * compile: ./make
 * run: ./part2
*)
open Image
open Cs1101 (* defines helpers: i2F, i2S, f2I, f2S, randomColor, ... *)
open World
       
(* constants *)

let displayWidth  = 800.                (* window width *)
let displayHeight = 800.                (* window height *)

(* type of world
 *)
type world_t = unit

(* Helpers *******************************************************)

let square side color = Image.rectangle side side color

let empty_scene width height =
  rectangle width height Color.transparent

let emptyImage = empty_scene displayWidth displayHeight

(* Parts of the problem set. **************************************** *)

type part = A | B | C | D | E | F | G



(* Part A: Picking A Color ****************************************** *)

(* Write a function 

  twoColors : unit -> Color.t

  such that a call (twoColors()) has equal chances of returning either
  Color.black or a randomly generated color.
 *)
let twoColors () =
  match (Random.float 1.0) < 0.5 with
  | true  -> Color.black
  | false -> randomColor()
			
let drawTwoColors () =
  let color = twoColors ()
  in
  square displayWidth color

let testTwoColors () =
  World.big_bang ()
		 ~name:"Tessellation Part A, Close this window to see Part B"
		 ~to_draw:drawTwoColors
		 ~width:(f2I displayWidth)
		 ~height:(f2I displayHeight)
  

(* Part B: RandomStripe ******************************************** *)

(* Write a function 

  randomStripe : unit -> Image.t

  such that a call (randomStripe()) returns a square the size of the
  graphics window with a horizontal stripe that is half height of the
  graphics window. The stripe should be centered and should have a 50%
  chance of being Color.black and 50% chance of being a random color.
 *)
let randomStripe () =
  let color = twoColors () in
  let height = displayHeight /. 2. in
  let myRectangle = rectangle displayWidth height color
  in
  place_image myRectangle (0., height /. 2.) emptyImage
			      
let testRandomStripe () =
  World.big_bang ()
		 ~name:"Tessellation Part B"
		 ~to_draw:randomStripe
		 ~width:(f2I displayWidth)
		 ~height:(f2I displayHeight)

(* Part C: Ring ***************************************************** *)

(* Write a function

  ring : float -> float -> Color.t -> Image.t

  such that a call (ring radius width color) returns a ring with a
  radius, width and color as given.
 *)
let ring radius width color =
  let outerCircle = (Image.circle radius color) in
  let innerRadius = radius -. width in
  let innerCircle = (Image.circle innerRadius Color.white)
  in
  Image.place_image innerCircle (width, width) outerCircle

let drawRing () =
  let radius = 300.0 in
  let ringWidth = 200.0 in
  let color = randomColor() in
  let (x, y) = (displayWidth /. 2. -. radius,
		displayHeight /. 2. -. radius)
  in
  place_image (ring radius ringWidth color) (x, y) emptyImage

let testRing () =
  World.big_bang ()
		 ~name:"Tessellation Part C"
		 ~to_draw:drawRing
		 ~width:(f2I displayWidth)
		 ~height:(f2I displayHeight)


(* Part D: Random Rings ********************************************* *)

(* Write a function

  randomRing : Image.t -> Image.t

  such that a call (randomRing image) returns an image that is just
  like the input image but it contains one newly placed random ring.
  By random, we mean, random radius, random width, random color and
  placed at random (x, y) coordinates on the input image.
 *)

let randomRing image =
  let radius = Random.float (displayWidth /. 2.) in
  let width  = Random.float radius in
  let color  = randomColor() in
  let x = Random.float displayWidth  -. 100. in
  let y = Random.float displayHeight -. 100.
  in
  place_image (ring radius width color)  (x, y) image

let changeWorldOnTick (n, image) = World (n - 1, randomRing image)
let drawRandomRings (n, image) = image
let finished (n, _) = n = 0
let drawFinal (_, image) = image

let testRandomRings () =
  let n = Random.int 500
  in
  World.big_bang (n, emptyImage)
		 ~name:"Tessellation Part D"
		 ~to_draw:drawRandomRings
		 ~on_tick:changeWorldOnTick
		 ~width:(f2I displayWidth)
		 ~height:(f2I displayHeight)
		 ~rate:0.05
		 ~stop_when:finished
		 ~to_draw_last:drawFinal


(* Part E: Row of Rings ********************************************* *)

(* Write a function

  rowOfRings : int -> Image.t

  such that a call (rowOfRings n) returns a rectangle containing n
  rings with equal diameters and with random widths and random
  colors. The height of the rectangle should be the diameter of the
  rings.  
*)
let rowOfRings n =
  let diameter = displayWidth /. (i2F n) in
  let radius = diameter /. 2. in
  let emptyBox = empty_scene displayWidth diameter in
  let rec repeat n image =
    match n with
    | 0 -> image
    | _ ->
      let ringWidth = Random.float radius in
      let color = randomColor() in
      let x = (i2F (n - 1)) *. diameter (* +. radius *) in
      let theRing = ring radius ringWidth color in
      let newImage = place_image theRing (x (* -. radius*), 0.) image
      in
      repeat (n - 1) newImage
  in
  repeat n emptyBox

let drawRowOfRings world =
  let n = (1 + Random.int 8) in
  let diameter = displayHeight /. (i2F n) in
  let radius = diameter /. 2. in
  let (x, y) = (0.0, displayHeight /. 2. -. radius)
  in
  place_image (rowOfRings n) (x, y) (empty_scene displayWidth diameter)

let testRowOfRings () =
  World.big_bang ()
		 ~name:"Tessellation Part E"
		 ~to_draw:drawRowOfRings
		 ~width:(f2I displayWidth)
		 ~height:(f2I displayHeight)


(* Part F: Diamonds ************************************************* *)

(* Write a function

  diamonds : int -> Image.t

  such that a call (diamonds n) returns an image containing n randomly
  colored squares of equal size running from upper left to lower right.
 *)
let diamonds n =
  let side = displayHeight /. (i2F n) in
  let rec repeat n =
    match n with
      0 -> (empty_scene displayWidth displayHeight)
    | _ ->
       let color = randomColor() in
       let sqr = square side color in
       let xy = (i2F (n - 1)) *. side
       in
       place_image (square side color) (xy, xy) (repeat (n - 1))
    in
    repeat n

let drawDiamonds world =
  let n = Random.int 8
  in
  diamonds n

let testDiamonds () =
  World.big_bang ()
		 ~name:"Tessellation Part F"
		 ~to_draw:drawDiamonds
		 ~width:(f2I displayWidth)
		 ~height:(f2I displayHeight)


(* Part G: Tessellation ********************************************* *)

let ringTessellation n =
  let diameter = displayWidth /. (i2F n) in
  let rec repeat m =
    match m with
      0 -> emptyImage
    | _ -> place_image (rowOfRings n) (0., (i2F (m - 1)) *. diameter) (repeat (m - 1))
  in
  repeat n

(* Squares Tessellation ********************************************* *)

let nestedSquare side width color =
  let innerSide = (side -. 2. *. width) in
  place_image (square innerSide Color.white) (width, width) (square side color)

(* A single nested square in the middle.
 *)
let testNestedSquare world =
  let color = randomColor()
  in
  place_image (nestedSquare 400. 100. color) (200., 200.) emptyImage

let rowOfSquares n =
  let side = displayWidth /. (i2F n) in
  let rec repeat n =
    match n with
      0 -> emptyImage
    | _ ->
      let squareWidth = Random.float side in
      let color = randomColor () in
      let x = (i2F (n - 1)) *. side +. (side /. 2.) in
      place_image (nestedSquare side squareWidth color) (x -. side /. 2., 0.) (repeat (n - 1))
  in
  repeat n;;

(* A row of nested squares across the middle.
 *)
let testRowOfSquares world =
  let n = Random.int 8 in
  let h = displayHeight /. (i2F n) in
  let (x, y) = (0., displayHeight /. 2. -. h /. 2.) in
  place_image (rowOfSquares n) (x, y) (empty_scene displayWidth h)

let squareTessellation n =
  let side = displayWidth /. (i2F n) in
  let rec repeat m =
    match m with
      0 -> (empty_scene displayWidth displayHeight)
    | _ -> place_image (rowOfSquares n) (0., (i2F (m - 1)) *. side) (repeat (m - 1))
  in
  repeat n

(* Hexagons ********************************************************)

let eqlTriangleHeight side =
  sqrt((side ** 2.0) -. (side /. 2.0) ** 2.0)

let hexagon side color =
  let a = side in (* using a, b and c for sides of right triangle. *)
  let b = eqlTriangleHeight side in
  let c = a /. 2. in
  let x2, y2 = (a, 0.) in
  let x3, y3 = (c, b) in
  let x4, y4 = (-.c, b) in
  let x5, y5 = (-.a, 0.) in
  let x6, y6 = (-.c, -.b) in
  let x1, y1 = (c, -.b) in
  let xys = [(x2, y2); (x3, y3); (x4, y4);
	     (x5, y5); (x6, y6); (x1, y1)]
  in
  (polygon xys color)

let hexagonDEBUG side color =
  let a = eqlTriangleHeight side in
  let (x, y) = (side, a) in
  let h = side /. 2. in
  let x1, y1 = (400., 0.) in
  let x2, y2 = (0., 400.) in
  let x3, y3 = (-400., 0.) in
  let x4, y4 = (0., -400.) in
  let xys = [(x1, y1); (x2, y2); (x3, y3); (x4, y4)] in
  let img0 = empty_scene 800. 800. in
  let img1 = empty_scene 400. 400. in
  let myLine = (line [(200., 200.)] ~size:10. Color.orange) in
  let img2 = place_image myLine (0., 0.) img1 in
  let img3 = place_image img2 (200., 200.) img0 in
  let myPoly = polygon xys ~fill:false ~outline_size:20. color
  in
  img3
(*  place_image myPoly (0., 0.) img1 *)


(* A single hexagon in the middle.
*)
let testHexagon world = hexagon 100. Color.white (*
  let side = 100. in
  let a = eqlTriangleHeight side in
  let empty = empty_scene displayWidth displayHeight in
  let (x, y) = (displayWidth /. 2. -. side, displayHeight /. 2. -. a)
  in
  place_image (hexagon side (randomColor())) (x, y) empty
						  *)
let rowOfHexagons n =
  let side = displayWidth /. (i2F n) /. 2. in
  let rec repeat n =
    match n with
      0 -> empty_scene displayWidth ((eqlTriangleHeight side) *. 2.)
    | _ ->
       let color = randomColor () in
       let x = (i2F (n - 1)) *. (side *. 3.)
       in
       place_image (hexagon side color) (x, 0.) (repeat (n - 1))
  in
  repeat n

(* a row of hexagons across the middle.
*)
let testRowOfHexagons world =
  let n = Random.int 8 in
  let side = displayWidth /. (i2F n) /. 3. in
  let a = eqlTriangleHeight side in
  let empty = empty_scene displayWidth displayHeight in
  let (x, y) = (0., displayHeight /. 2. -. a)
  in
  place_image (rowOfHexagons n) (x, y) empty

let hexTessellation n =
  let side = displayWidth /. (i2F n) /. 2. in
  let dy = eqlTriangleHeight side in
  let rec repeat m y =
    match y > displayHeight with
      true  -> (empty_scene displayWidth displayHeight)
    | false ->
      let x = (match m mod 2 with
		 0 -> -.(side /. 2.)
               | _ -> side) in
      place_image (rowOfHexagons n) (x, y) (repeat (m + 1) (y +. dy))
  in
  repeat 0 (-.dy)


(* Chickenwire ***************************************************)

let chickenWire side thickness color =
  let innerSide = side -. thickness in
  let hex = hexagon side color in
  place_image (hexagon innerSide Color.blue) (thickness, thickness) hex

let rowOfChickenWire n =
  let side = displayWidth /. (i2F n) /. 2. in
  let rec repeat n =
    match n with
      0 -> empty_scene displayWidth ((eqlTriangleHeight side) *. 2.)
    | _ ->
      let color = randomColor () in
      let x = (i2F (n - 1)) *. (side *. 3.) in
      let thickness = 10. in
      place_image (chickenWire side thickness color) (x, 0.) (repeat (n - 1))
  in
  repeat n

let chickenWireTessellation n =
  let side = displayWidth /. (i2F n) /. 2. in
  let dy = eqlTriangleHeight side in
  let rec repeat m y =
    match y > displayHeight with
      true  -> (empty_scene displayWidth displayHeight)
    | false ->
      let x = (match m mod 2 with
		 0 -> -.(side /. 2.)
               | _ -> side)
      in
      place_image (rowOfChickenWire n) (x, y) (repeat (m + 1) (y +. dy))
  in
  repeat 0 (-.dy)

(* ***************************************************************** *)

(* This test oscillates between ring, square and hexagonal tessellations. *)
	 
let variousTessellations world =
  let n = (1 + Random.int 7)
  in
  match (Random.int 4) mod 4 with 
    0 -> ringTessellation n
  | 1 -> squareTessellation n
  | 2 -> hexTessellation n
  | 3 -> testHexagon () 
  | _ -> chickenWireTessellation n

let testTessellation world =
  World.big_bang ()
		 ~name:"Tessellation Part G"
		 ~to_draw:variousTessellations
		 ~width:(f2I displayWidth)
		 ~height:(f2I displayHeight)

(*********************************************************************)

 let tryPart part =
   match part with
   | A -> testTwoColors () 
   | B -> testRandomStripe()
   | C -> testRing ()
   | D -> testRandomRings ()
   | E -> testRowOfRings ()
   | F -> testDiamonds ()
   | G -> testTessellation () 

let _ = tryPart A
let _ = tryPart B
let _ = tryPart C
let _ = tryPart D
let _ = tryPart E
let _ = tryPart F
let _ = tryPart G

